unit module;

{$mode objfpc}

interface

procedure init;
function getN : integer;
function getM : integer;
function getP : integer;
procedure getView(num : integer; const mirrors, states : array of integer; var romeo_view, juliet_view : integer);

{ ne e testvano!!!
  osven tova v nachaloto si pomislih che mirrors i states sa promenlivi, koito
  se vryshtat ot getView na user-a, a ne ukazatel kym masiv kakto qvno e.
  promenih go s open array parameter.
  (deklaraciqta 'array of integer' e open array parameter)
  
  na nego mu se predava skrit parametyr - naj-golemiq
  index na masiva: high(mirrors).
  naj-malkiq index e vinagi 0.
  tova znachi che mozhe bi num e izlishno.

  tova mirrors kakvo e? trqbva li da e const vmesto var? t.e. user-a li go
  predava na modula ili modula vryshta neshto tam? vyobshte bez da znam
  zadachata ne e mnogo lesno da precenq koj e naj-dobriq nachin
}

implementation

uses sysutils;

const
  TEST_FILE   = 'crazytra.in';
  OUT_FILE	  = 'crazytra.out';	
  STATE_DOWN  = 0;
  STATE_RIGHT = 1;
  STATE_LEFT  = 2;
  MAX_SIZE    = 128;
  MAX_P       = 128;
  DIR_UP      = 0;
  DIR_DOWN    = 1;
  DIR_LEFT    = 2;
  DIR_RIGHT   = 3;
  ROMEO       = 0;
  JULIET      = 1;

var
  init_called : boolean;
  n, m, p : integer;
  romeo_pos, romeo_row, romeo_col, romeo_dir : integer;
  juliet_pos, juliet_row, juliet_col, juliet_dir : integer;
  state_table   : array[0..MAX_SIZE-1, 0..MAX_SIZE-1] of integer;
  mirror_pos    : array[0..MAX_P-1, 0..1] of integer;
  position_used : array[0..MAX_SIZE-1, 0..MAX_SIZE-1] of byte; // one based index of a changeable mirror; zero otherwise

  path_number : array[0..MAX_P+3, 0..2, 0..3, 0..MAX_P-1] of integer; // mirror; state; direction; mirror position on path
  path_dir    : array[0..MAX_P+3, 0..2, 0..3, 0..MAX_P-1] of integer;
  path_end    : array[0..MAX_P+3, 0..2, 0..3] of integer;

// should be synchronized with DIR_* and STATE_* definitions
const
  transition_table : array[0..3, 0..2] of integer =
(
  ( DIR_UP,    DIR_RIGHT, DIR_LEFT  ),
  ( DIR_DOWN,  DIR_LEFT,  DIR_RIGHT ),
  ( DIR_LEFT,  DIR_DOWN,  DIR_UP    ),
  ( DIR_RIGHT, DIR_UP,    DIR_DOWN  )
);

// should be synchronized with DIR_* definitions
  row_change : array[0..3] of integer = ( -1, 1, 0, 0);
  col_change : array[0..3] of integer = ( 0, 0, -1, 1);

var
  fp : TextFile;
  fpo : TextFile;
  fpIsOpen : Boolean;

// exits with an error message
procedure my_exit(const msg : string);
begin
  //assign(fpo, OUT_FILE);
  //rewrite(fpo);
  writeln(msg);
  //close(fpo);

  if fpIsOpen then
    CloseFile(fp);

  halt(0);
end;

// help function for getting the information out of a position around the table
procedure getDetailedInfo(pos : integer; var r, c, dir : integer);
begin
  if pos <= n then
  begin
    r := pos-1; c := 0; dir := DIR_RIGHT;
  end
  else
    if pos <= n+m then
    begin
      r := n-1; c := pos-n-1; dir := DIR_UP;
    end
    else
      if pos <= n+m+n then
      begin
        r := n-(pos-n-m); c := m-1; dir := DIR_LEFT;
      end
      else
      begin
        r := 0; c := m-(pos-n-m-n); dir := DIR_DOWN;
      end;
end;

// returns position around the table based on row and column information
function getCompiledPosition(r, c : integer) : integer;
begin
  if r = -1 then // top side
    result := n+m+n+m-c
  else
  if c = -1 then // left side
    result := r+1
  else
  if r = n then  // bottom side
    result := n+c+1
  else
  if c = m then  // right side
    result := n+m+n-r;
end;

// precomputes data for a single changeable mirror
// note : function works for romeo and juliet too
procedure mirror_precompute(ind, r, c, dir, current_state : integer);

var
  start_r, start_c, start_dir : integer;
  state, current_number, mirr : integer;

begin
  start_r := r; start_c := c; start_dir := dir;
  current_number := 1;

  if (ind>=p) and (position_used[r, c]<>0) then // romeo and juliet may encounter changeable mirror on the first position of their path
  begin
    mirr := position_used[r, c] - 1;
    path_number[ind, current_state, dir, mirr] := current_number;
    path_dir[ind, current_state, dir, mirr] := dir;
  end;

  while (r>=0) and (r<n) and (c>=0) and (c<m) do
  begin
    state := state_table[r, c];

    if (ind<p) and (r=mirror_pos[ind, 0]) and (c=mirror_pos[ind, 1]) then
      state := current_state;

    dir := transition_table[dir, state];
    inc(r, row_change[dir]);
    inc(c, col_change[dir]);

    inc(current_number);

    if (r=start_r) and (c=start_c) and (dir=start_dir) then // cycle found : no need to save anything since it won't be used in the query function
      exit;

    if (r>=0) and (r<n) and (c>=0) and (c<m) then
      if position_used[r, c] <> 0 then
      begin
        mirr := position_used[r, c] - 1;
        if path_number[ind, current_state, start_dir, mirr] = 0 then // some mirrors may appear more than once; ignore all appearances after the first one
        begin
          path_number[ind, current_state, start_dir, mirr] := current_number;
          path_dir[ind, current_state, start_dir, mirr] := dir;
end;
      end;
  end;

  path_end[ind, current_state, start_dir] := getCompiledPosition(r, c);
end;

// precomputes data so that queries can be done fast
procedure precompute_path_data;

var
  i, j, k : integer;

begin
  for i:=0 to p-1 do // for each mirror
    for j:=0 to 2 do // for each state
      for k:=0 to 3 do // for each direction
        mirror_precompute(i, mirror_pos[i][0], mirror_pos[i][1], k, j);

  mirror_precompute(p, romeo_row, romeo_col, romeo_dir, 0);      // for romeo
  mirror_precompute(p+1, juliet_row, juliet_col, juliet_dir, 0); // for juliet
end;

// call before any other calls
procedure init;

var
  i, j : integer;

begin

  if init_called then
     my_exit('You're trying to call init more than once!');

  init_called := true;

  AssignFile(fp, TEST_FILE);
  Reset(fp);
  fpIsOpen := true;

  if IOResult <> 0 then
    my_exit('Input file does not exist!');

  try
    read(fp, n, m);
  except
    my_exit('Invalid input file!');
  end;

  if (n<1) or (n>MAX_SIZE) or (m<1) or (m>MAX_SIZE) then
    my_exit('Invalid table dimensions!');

  for i:=0 to n-1 do
    for j:=0 to m-1 do
    begin
      try
        read(fp, state_table[i, j]);
      except
        my_exit('Invalid input file!');
      end;

      if (state_table[i,j]<>STATE_DOWN) and (state_table[i,j]<>STATE_RIGHT) and (state_table[i,j]<>STATE_LEFT) then
        my_exit('Some of the table cells are invalid!');
    end;


  try
    read(fp, romeo_pos, juliet_pos);
  except
    my_exit('Invalid input file!');
  end;

  if (romeo_pos<1) or (romeo_pos>m+m+n+n) or (juliet_pos<1) or (juliet_pos>m+m+n+n) then
    my_exit('Romeo and/or Juliet position is invalid!');


  try
    read(fp, p);
  except
    my_exit('Invalid input file!');
  end;

  if (p<0) or (p>MAX_P) then
    my_exit('Number of changeable mirrors is invalid!');

  for i:=0 to p-1 do
  begin
    try
      read(fp, mirror_pos[i, 0], mirror_pos[i, 1]);
    except
      my_exit('Invalid input file!');
    end;

    if (mirror_pos[i, 0]<0) or (mirror_pos[i, 0]>=n) or (mirror_pos[i, 1]<0) or (mirror_pos[i, 1]>=m) then
      my_exit('Invalid mirror positions!');

    if position_used[mirror_pos[i, 0], mirror_pos[i, 1]] <> 0 then
      my_exit('Some of the changeable mirrors have the same position!');

    position_used[mirror_pos[i, 0], mirror_pos[i, 1]] := i+1;
  end;


  fpIsOpen := False;
  CloseFile(fp);

  getDetailedInfo(romeo_pos, romeo_row, romeo_col, romeo_dir);
  getDetailedInfo(juliet_pos, juliet_row, juliet_col, juliet_dir);

  precompute_path_data;
end;

// returns table height
function getN : integer;
begin
  if not init_called then
    my_exit('Init not called!');

  result := n;
end;

// returns table width
function getM : integer;
begin
  if not init_called then
    my_exit('Init not called!');

  result := m;
end;

// returns number of moveable mirrors
function getP : integer;
begin
  if not init_called then
    my_exit('Init not called!');

  result := p;
end;

var
  listed_mirrors : array[0..MAX_P-1] of boolean;

// checks the input of the user and exits with an error message if something is wrong
// input for the getView function
procedure check_input(num : integer; const mirrors, states : array of integer);

var
  i : integer;

begin
  if (num<0) or (num>6) then
    my_exit('Query with an invalid number of mirrors!');

  for i:=0 to num-1 do
  begin
    if (mirrors[i]<0) or (mirrors[i]>=p) then
      my_exit('Query with an invalid mirror number!');

    if (states[i]<>STATE_DOWN) and (states[i]<>STATE_RIGHT) and (states[i]<>STATE_LEFT) then
      my_exit('Query with an invalid mirror state!');

    if listed_mirrors[mirrors[i]] then
      my_exit('Some mirrors are repeated in the list!');

    listed_mirrors[mirrors[i]]:=true;
  end;

  for i:=0 to num-1 do
    listed_mirrors[mirrors[i]]:=false;
end;

// not a public function
// returns ray destination as a number around the table
// this version of the module uses this version of the function only to check the faster one for correctness
function getFinalDestination(r, c, dir, num : integer; const mirrors, states : array of integer) : integer;

var
  i, state : integer;

begin
  while (r>=0) and (r<n) and (c>=0) and (c<m) do
  begin
    state := state_table[r, c];

    for i:=0 to num-1 do
      if (mirror_pos[mirrors[i], 0]=r) and (mirror_pos[mirrors[i], 1]=c) then
        state := states[i];

    dir := transition_table[dir, state];
    inc(r, row_change[dir]);
    inc(c, col_change[dir]);
  end;

  result := getCompiledPosition(r, c);
end;

// not a public function
// returns ray destination as a number around the table
// faster version of the above function
// uses precomputing and has complexity somewhere around num^2
function getFinalDestinationFast(person, r, c, dir, num : integer; const mirrors, states : array of integer) : integer;

var
  i, state, ind, best, bestind, tmp : integer;

begin
  if person = ROMEO then
    ind := p
  else
    ind := p+1;

  state := 0; // data for romeo and juliet is written as state 0 though they don't have a state as mirrors do

  while true do
  begin
    best := high(integer);
    for i:=0 to num-1 do
    begin
      tmp := path_number[ind, state, dir, mirrors[i]];
      if (tmp<>0) and (tmp<best) then
      begin
        best:=tmp;
        bestind:=i;
      end;
    end;

    if best = high(integer) then // no changed mirror found on the path
    begin
      result := path_end[ind, state, dir];
      exit;
    end;

    dir:=path_dir[ind, state, dir, mirrors[bestind]];
    state:=states[bestind];
    ind:=mirrors[bestind];
  end;

  result := -1;
end;

// used by the contester to check his/her guesses
procedure getView(num : integer; const mirrors, states : array of integer; var romeo_view, juliet_view : integer);

var
  r2, j2 : integer;

begin
  if not init_called then
    my_exit('Init not called!');

  check_input(num, mirrors, states);

  romeo_view := getFinalDestinationFast(ROMEO, romeo_row, romeo_col, romeo_dir, num, mirrors, states);
  juliet_view := getFinalDestinationFast(JULIET, juliet_row, juliet_col, juliet_dir, num, mirrors, states);

  if (romeo_view=juliet_pos) and (juliet_view=romeo_pos) then
    my_exit('Mnogo tajno syobshtenie.');

end;

end.
